#ifndef Models_H
#define Models_H

#include <stdio.h>
#include <math.h>
#include <stdlib.h>

/* function prototypes */

/*
c---------------------------------------------------------------------
c                         Black-Scholes Model
c---------------------------------------------------------------------
c     Routine name           - bscall 
c---------------------------------------------------------------------
c         Purpose            - This subroutine calculates the value of
c                              european call option using the formula
c                              derived by black and scholes for a
c                              payout protected european call. no
c                              adjustment is made for dividends.
c                              (black-scholes model)
c---------------------------------------------------------------------
c     Arguments: 
c        sp --- initial stock price  sk --- exercise (striking) price
c        r  --- risk-free rate       t  --- time to maturity
c        v  --- initial volatility   cval --- call price (output)
c---------------------------------------------------------------------
c     Required routines - pnorml
c---------------------------------------------------------------------
*/
float bscall(sp,sk,r,t, v)
     float sp,sk,r,t,v;
{
  float hedge,h1,h2,p,tmp1,tmp2, cval,pnorml();
 
  /*c     --- calculate black-scholes call value */
  tmp1 = sk/exp(r*t);
  tmp2 = v*sqrt(t);
  /*c     --- h = (log(sp/(sk/(exp(r*t)))))/(v*sqrt(t))+.5*(v*sqrt(t)) */
  h1 = log(sp/tmp1)/tmp2 + 0.5*tmp2;
  h2 = h1 - tmp2;
  hedge = pnorml(h1);
  p =  pnorml(h2);

  cval = sp*hedge - tmp1*p;

  return (cval);
}

/*
c---------------------------------------------------------------------    
c     Routine name         - pnorml (used by Black-Scholes Model)         
c---------------------------------------------------------------------
c         Purpose          - This function supplies a polynomial
c                            approximation for the standard normal
c                            distribution.
c---------------------------------------------------------------------
c     Required routines - none
c---------------------------------------------------------------------
*/
float pnorml(z)
      float z;
{

  float tmp1,tmp2,tmp3,tmp4;
  
  tmp1 = z/sqrt(2.0);
  tmp2 = fabs(tmp1);
  tmp3 = 1./(1. + 0.5*tmp2);
  tmp4 = tmp3*exp(-tmp2*tmp2 - 1.26551223 + tmp3*(1.00002368 + 
	 tmp3*(.37409196 + tmp3*(.09678418 + tmp3*(-.18628806 +
	 tmp3*(.27886807 + tmp3*(-1.13520398 + tmp3*(1.48851587 + 
	 tmp3*(-.82215223 + tmp3*.17087277)))))))));

  if (tmp1<0.0) tmp4 = 2. - tmp4;
  
  return (1. - 0.5 * tmp4);
}

#define nmax 24

float implied_vol_bs(s,x,r,t,c,div,tdiv)
/* Use bisection method to find volatility */
float s,x,r,t,c,div,tdiv;
{
  float v1,v2,v0,vmid,fmid,dv,f,sdiv;
  int i;

  sdiv = s - div*exp(-r*tdiv);
  v1 = 0.0000006; v2 = 1.2;
  fmid = (bscall(sdiv,x,r,t,v2) - c);
  f = (bscall(sdiv,x,r,t,v1) - c);
  if (f*fmid >= 0.0) 
    {
      /* fprintf(stderr,"implied_vol: root not bracketed at %i\n",count);*/
      return(-1.0);
    }

/*  printf("f, fmid = %lf, %lf\n",f,fmid);*/
  if (f < 0.0) 
    {
      v0 = v1;
      dv = v2 - v1;
    }
  else
    {
      v0 = v2;
      dv = v1 - v2;
    }
  for (i = 0; i < nmax; i++)
    {
      dv = dv/2.0;
      vmid = v0 + dv;
      fmid = (bscall(sdiv,x,r,t,vmid) - c);
      if (fmid < 0.0) v0 = vmid;
/*      printf("%i   %lf  %lf  %lf\n",i,v0,dv,fmid);*/
    }

  return(v0);
}

#endif Models_H
